from typing import Callable

from ..checks import assert_str


class InvalidArgumentType(Exception):
    ...


class CheckIn:
    def __init__(self, name, check_fn, description=None, param_name=None):
        assert_str(name, "CheckIn.__init__:name")
        assert_str(name, "CheckIn.__init__:description")

        self.name = name
        self.check_fn = check_fn
        self.description = description
        self.param_name = param_name

    def add_param_name(self, param_name: str):
        self.param_name = param_name

    def run_check(self, val):
        try:
            possible_bool_result = self.check_fn(val)

            if possible_bool_result is not None and possible_bool_result != True:
                raise InvalidArgumentType(
                    f"Failed running check '{self.name}' for parameter '{self.param_name}'. Got value '{val}'."
                )

        except Exception as check_error:
            raise


class _In:
    def __init__(self, param_name, checks):
        self.param_name = param_name
        self.check_ins = []

        if isinstance(checks, Callable):
            self.check_ins.append(
                CheckIn(
                    name=f"Check {param_name}", check_fn=checks, param_name=param_name
                )
            )
        elif isinstance(checks, CheckIn):
            self.check_ins.append(checks)
        elif isinstance(checks, dict):
            for check_name, check in checks.items():
                self.check_ins.append(
                    CheckIn(name=check_name, check_fn=check, param_name=param_name)
                )
        elif isinstance(checks, list):
            for check in checks:
                if not isinstance(check, CheckIn):
                    raise Exception()

    def run_check(self, val):
        for check_in in self.check_ins:
            check_in.run_check(val)


class _InGroup:
    def __init__(self, ins_dict):
        self.ins = []

        if ins_dict is None:
            return

        for param_name, _in in ins_dict.items():
            self.ins.append(_In(param_name, _in))

    def run_check(self, ins):
        for check_in in self.ins:
            check_in.run_check(ins[check_in.param_name])



from rcheck.definitions.ins import _InGroup
from rcheck.definitions.out import _Out
from rcheck.definitions.post_check import _PostGroup
from rcheck.definitions.pre_check import _PreGroup


class InputValues:
    def __init__(self, input_dict):
        for input_name, input_value in input_dict.items():
            setattr(self, input_name, input_value)


class Check:
    def __init__(
        self, run_fn, ins=None, pre_check=None, out=None, post_check=None, off=False
    ):
        self.run_fn = run_fn

        # check ins is dict of ...
        self.ins = _InGroup(ins)

        self.pre_check = _PreGroup(pre_check)
        self.out = _Out(out)
        self.post_check = _PostGroup(post_check)
        self.off = off

    def __call__(self, *args, **kwargs):
        if self.off:
            return self.call_without_checking(*args, **kwargs)

        inputs_as_dict = kwargs

        self.ins.run_check(ins=inputs_as_dict)
        input_class = InputValues(kwargs)

        self.pre_check.run_check(input_class)

        result = self.run_fn(*args, **kwargs)
        self.out.run_check(result)

        self.post_check.run_check(input_class, result)

        return result

    def call_without_checking(self, *args, **kwargs):
        return self.run_fn(*args, **kwargs)



from typing import Optional

from rcheck.checks.basic_types import _assert_simple_type_fail_message
from rcheck.checks.shared import InvalidRuntype, _isinstance

#
# list
#


def _assert_list_type_fail_message(
    expected_type_str: str,
    val: object,
    name: str,
    failure_index: int,
    message: Optional[str] = None,
):
    return InvalidRuntype(
        f"Expected list items of {name} to be type {expected_type_str}. "
        + f"Got value {val!r} of type {type(val).__name__} at index {failure_index}."
        + (f"\n\n{message}" or "")
    )


def _is_iterable_of(iterable, sub_type):
    for idx, item in enumerate(iterable):
        if not _isinstance(item, sub_type):
            return False, idx

    return True, -1


def is_list(val: object, of: Optional[type] = None):
    if not _isinstance(val, list):
        return False

    return True if of is None else _is_iterable_of(val, of)[0]


def assert_list(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    of: Optional[type] = None,
):
    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    if of is None:
        return

    items_pass_check, failed_idx = _is_iterable_of(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            of.__name__, val[failed_idx], name, failed_idx, message
        )


def _is_iterable_of_opt(iterable, sub_type):
    for idx, item in enumerate(iterable):
        if item is not None and not _isinstance(item, sub_type):
            return False, idx

    return True, -1


def is_list_of_opt(val: object, of_optional: type):
    if not _isinstance(val, list):
        return False

    return _is_iterable_of_opt(val, of_optional)[0]


def assert_list_of_opt(
    val: object,
    of_optional: type,
    name: str,
    message: Optional[str] = None,
):
    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    if of_optional is None:
        return

    items_pass_check, failed_idx = _is_iterable_of_opt(val, of_optional)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            of_optional.__name__, val[failed_idx], name, failed_idx, message
        )


def is_opt_list(val: object, of: Optional[type] = None):
    return val is None or is_list(val, of)


def assert_opt_list(
    val: object,
    name: str,
    message: Optional[str] = None,
    *,
    of: Optional[type] = None,
):
    if val is None:
        return

    if not _isinstance(val, list):
        raise _assert_simple_type_fail_message("list", val, name, message)

    if of is None:
        return

    items_pass_check, failed_idx = _is_iterable_of(val, of)
    if not items_pass_check:
        raise _assert_list_type_fail_message(
            of.__name__, val[failed_idx], name, failed_idx, message
        )


#
# dict
#



from typing import Callable

from rcheck.checks import assert_opt_str, assert_str


class InvalidOutType(Exception):
    ...


class CheckOut:
    def __init__(self, name, check_fn, description=None):
        assert_str(name, "name")
        assert_opt_str(description, "description")

        self.name = name
        self.check_fn = check_fn
        self.description = description

    def run_check(self, val):
        try:
            possible_bool_result = self.check_fn(val)

            if possible_bool_result is not None and possible_bool_result != True:
                raise InvalidOutType(
                    f"Failed running check '{self.name}'. Got value '{val}':{type(val).__name__}."
                )

        except Exception as check_error:
            raise


class _Out:
    def __init__(self, outs_dict):
        self.outs = []

        if outs_dict is None:
            return

        for check_name, out in outs_dict.items():
            if isinstance(out, Callable):
                self.outs.append(CheckOut(check_name, out))
            elif isinstance(out, CheckOut):
                self.outs.append(out)

    def run_check(self, out):
        for check_out in self.outs:
            check_out.run_check(out)



from typing import Optional

from .checks import assert_opt_bool
from .definitions import Check


def check(
    run_fn=None,
    *,
    ins=None,
    pre_check=None,
    out=None,
    post_check=None,
    off: Optional[bool] = None
):
    """

    Params
    ------
    run_fn
    ins
    pre_check
    out
    post_check

    """

    # If this function is used:
    #
    #   1. Without passing in any arguments. E.g.:
    #      >>> @check
    #      ... def some_func():
    #      ...     ...
    #
    #   2. Not as a decorator function. E.g.:
    #      >>> check(some_func, ...)
    #

    if run_fn is not None:
        fn_name = ...

        assert_opt_bool(off, "off")

        if ins is None:
            # ins from signature typing
            ...

        if out is None:
            # out from signature typing
            ...

        return Check(
            run_fn=run_fn,
            ins=ins,
            pre_check=pre_check,
            out=out,
            post_check=post_check,
            off=off,
        )

    # If this function is used as a decorator with arguments:
    #
    # E.g:
    # >>> @check(ins=..., ...)
    # ... def some_func():
    # ...     ...
    def _check(run_fn):
        return check(
            run_fn,
            ins=ins,
            pre_check=pre_check,
            out=out,
            post_check=post_check,
            off=off,
        )

    return _check



class FailedPostCheck(Exception):
    ...


class _PostCheck:
    def __init__(self, check_name, check_fn):
        self.check_name = check_name
        self.check_fn = check_fn

    def run_check(self, ins, out):
        try:
            possible_bool_result = self.check_fn(ins, out)

            if possible_bool_result is not None and possible_bool_result != True:
                raise FailedPostCheck(f"Failed running check '{self.check_name}'")

        except Exception as check_error:
            raise


class _PostGroup:
    def __init__(self, post_checks_dict):
        self.post_checks = []

        if post_checks_dict is None:
            return

        for param_name, _in in post_checks_dict.items():
            self.post_checks.append(_PostCheck(param_name, _in))

    def run_check(self, ins, out):
        for post_check in self.post_checks:
            post_check.run_check(ins, out)
